
import * as sch from '../utilities/json-schema';
import { FastifyInstance, RouteShorthandOptions } from 'fastify';
import { app_name, app_version } from '../conf';
import { hostname } from 'os';
import { simple_memo } from '../utilities/memo';
import { StorageProvider, StorageStatus } from '../storage';
import { HttpConfig, ServerStatus } from '../http/server';

export interface Dependencies {
	ready_check: () => ServerStatus;
	storage: StorageProvider;
}

export const status_schema = sch.obj({
	status: sch.str_enum([ 'unstarted', 'starting', 'ready', 'closing', 'closed' ]),
	hostname: sch.str(),
	app_name: sch.str_enum([ app_name ]),
	app_version: sch.str_enum([ app_version ]),
	store: sch.obj({
		engine: sch.str(),
		status: sch.str_enum([ 'ok', 'updating', 'warning', 'unavailable' ])
	}, { additionalProperties: true }),
	readiness: sch.str('uri'),
});

export interface FullStatus {
	status: ServerStatus;
	hostname: string;
	app_name: string;
	app_version: string;
	store: StorageStatus;
}

export function register_status_endpoint(http_server: FastifyInstance, conf: HttpConfig, { ready_check, storage }: Dependencies) {
	const get_hostname = simple_memo(30000, hostname);
	
	const opts: RouteShorthandOptions = {
		schema: {
			response: {
				200: status_schema,
			},
		},
	};
	
	http_server.get('/status', opts, async (req, res) => {
		const content: FullStatus = {
			status: ready_check(),
			hostname: get_hostname(),
			app_name,
			app_version,
			store: await storage.status(),
		};

		return content;
	});
}
