
import * as sch from '../../utilities/json-schema';
import { render_login_page } from './login-page';
import { send_html_error } from '../../http/send-error';
import { redirect_200_refresh } from '../../http/redirects';

import type { HttpConfig } from '../../http/server';
import type { HttpWebDependencies } from '../server';
import type { HttpURL, Locale, Timezone } from '../../utilities/types';
import type { FastifyInstance, FastifyRequest, RouteShorthandOptions } from 'fastify';
import { UserinfoResponse } from 'openid-client';

export function register_login_callback_endpoint(http_server: FastifyInstance, conf: HttpConfig, deps: HttpWebDependencies) {
	const { logger, pkce_cookie, oidc, storage, snowflake, session, argon2 } = deps;
	const opts: RouteShorthandOptions = {
		schema: {
			response: {
				200: { },
				401: { },
			},
			querystring: sch.obj({
				code: sch.str(),
				state: sch.str(),
				session_state: sch.str(),
				error: sch.str(),
				error_description: sch.str(),
				error_uri: sch.str(),
			}, { additionalProperties: true })
		},
	};

	type Req = FastifyRequest<{
		Querystring: {
			code?: string;
			state?: string;
			session_state?: string;
			error?: string;
			error_description?: string;
			error_uri?: string;
		};
	}>;

	http_server.get('/login-callback', opts, async (req: Req, res) => {
		const log = logger.child({ reqId: req.id }, {
			redact: [
				'callback_params.access_token',
				'callback_params.code',
				'callback_params.id_token',
				'callback_params.session_state',
			],
		});

		const pkce_code_verifier = pkce_cookie.read_pkce_code_verifier(req);

		if (! pkce_code_verifier) {
			log.debug('no pkce code verifier provided');
			return send_html_error(res, 'oidc_no_pkce_code_verifier', render_login_page);
		}

		const params = oidc.parse_callback_params(req.url);
		
		if (! params) {
			return send_html_error(res, 'oidc_callback_params_invalid', render_login_page);
		}

		log.debug({ callback_params: params }, 'received callback params');

		const token_set = await oidc.fetch_token_set(`${conf.exposed_url}/login-callback`, params, pkce_code_verifier);
		
		if (! token_set) {
			return send_html_error(res, 'oidc_token_fetch_failed', render_login_page);
		}

		log.debug('fetched token set; requesting user info');

		const user_info = await oidc.fetch_user_info(token_set);

		if (! user_info) {
			return send_html_error(res, 'oidc_userinfo_fetch_failed', render_login_page);
		}

		log.debug({ oidc_subject: user_info.sub }, 'fetched user info; looking up local user data');
		log.debug({ user_info }, 'userinfo');

		let user = await storage.get_user_by_oidc_sub(user_info.sub);

		if (! user) {
			log.debug('user does not have a local profile; creating a new one');

			// todo: handle non-unique usernames
			user = {
				id: snowflake.uid_str(),
				oidc_subject: user_info.sub,
				username: user_info.preferred_username,
				name: user_info.name,
				locale: get_locale(user_info),
				timezone: get_zoneinfo(user_info),
				picture: user_info.picture as HttpURL,
				profile: user_info.profile as HttpURL,
				website: user_info.website as HttpURL,
			};
			
			await storage.create_user(user);
		}

		log.debug({ user_id: user.id }, 'login successful; creating new local session');

		const session_key = await session.generate_key();
		const session_data = await session.create_session_data(user.id, session_key);
		await storage.create_session(session_data);

		pkce_cookie.reset(res);
		session.write_to_cookie(res, session_key);
		redirect_200_refresh(res, conf.exposed_url, 'Login successful');
	});
}

function get_locale(user_info: UserinfoResponse) : Locale {
	const from_userinfo = user_info.locale?.replace('_', '-') as Locale;
	return from_userinfo || 'en-US';
}

function get_zoneinfo(user_info: UserinfoResponse) : Timezone {
	const from_userinfo = user_info.zoneinfo as Timezone;
	return from_userinfo || 'Africa/Abidjan';
}
