
import { DateTime } from 'luxon';
import { IncomingHttpHeaders } from 'http';

export interface ParsedCacheHeaders {
	age?: number;
	etag?: string;
	vary?: string[];
	date?: DateTime;
	cache_control?: CacheControl;
	expires?: DateTime;
	last_modified?: DateTime;
}

export interface CacheControl {
	private?: boolean;
	no_store?: boolean;
	no_cache?: boolean;
	max_age?: number;
	must_revalidate?: boolean;
	proxy_revalidate?: boolean;
}

export function parse_cache_headers(headers: IncomingHttpHeaders) {
	const result: ParsedCacheHeaders = { };

	if (headers['age']) {
		result.age = parseInt(headers['age'], 10);
	}

	if (headers['etag']) {
		result.etag = headers['etag'];
	}

	if (headers['vary']) {
		result.vary = headers['vary'].split(',').map((str) => str.trim().toLowerCase());
	}

	if (headers['date']) {
		result.date = DateTime.fromHTTP(headers['date']);
	}

	if (headers['cache-control']) {
		result.cache_control = { };

		for (let directive of headers['cache-control'].split(',')) {
			directive = directive.trim();

			switch (directive) {
				case 'private':
					result.cache_control.private = true;
					break;
				case 'no-store':
					result.cache_control.no_store = true;
					break;
				case 'no-cache':
					result.cache_control.no_cache = true;
					break;
				case 'must-revalidate':
					result.cache_control.must_revalidate = true;
					break;
				case 'proxy-revalidate':
					result.cache_control.proxy_revalidate = true;
					break;
				default:
					if (directive.startsWith('max-age=')) {
						result.cache_control.max_age = parseInt(directive.slice(8), 10);
						break;
					}

					// todo: log something here about unknown directive
			}
		}
	}

	if (headers['expires']) {
		result.expires = DateTime.fromHTTP(headers['expires']);
	}

	if (headers['last-modified']) {
		result.last_modified = DateTime.fromHTTP(headers['last-modified']);
	}

	return result;
}
