
import { FastifyInstance, FastifyRequest } from 'fastify';
import { XMLParser, XMLValidator, X2jOptions } from 'fast-xml-parser';
import * as yaml from 'yaml';

export function json_content_parser(http_server: FastifyInstance, media_types: string[]) {
	http_server.addContentTypeParser(media_types, { parseAs: 'string' }, http_server.getDefaultJsonParser('ignore', 'ignore'));
}

export function yaml_content_parser(http_server: FastifyInstance, media_types: string[]) {
	http_server.addContentTypeParser(media_types, { parseAs: 'string' }, yaml_content_processor);
}

export function xml_content_parser(http_server: FastifyInstance, media_types: string[]) {
	http_server.addContentTypeParser(media_types, { parseAs: 'string' }, xml_content_processor);
}

export function text_content_parser(http_server: FastifyInstance, media_types: string[]) {
	http_server.addContentTypeParser(media_types, { parseAs: 'string' }, noop_content_processor);
}

export function binary_content_parser(http_server: FastifyInstance, media_types: string[]) {
	http_server.addContentTypeParser(media_types, { parseAs: 'buffer' }, noop_content_processor);
}



function noop_content_processor(req: FastifyRequest, payload: string | Buffer, done: (err: Error | null, body?: any) => void) {
	done(null, payload);
}

function xml_content_processor(req: FastifyRequest, payload: string, done: (err: Error | null, body?: any) => void) {
	const opts: Partial<X2jOptions> = { };
	const xmlParser = new XMLParser(opts);
	const result = XMLValidator.validate(payload, opts);

	if (typeof result === 'object' && result.err) {
		done(new RequestContentParseError(result.err.msg));
		return;
	}

	done(null, xmlParser.parse(payload));
}

function yaml_content_processor(req: FastifyRequest, payload: string, done: (err: Error | null, body?: any) => void) {
	let result: unknown;

	try {
		result = yaml.parse(payload);
	}

	catch (error) {
		const message = error?.name === 'YAMLParseError' ? error.message : 'Failed to parse request body';
		done(new RequestContentParseError(message));
		return;
	}

	done(null, result);
}

export class RequestContentParseError extends Error {
	public readonly name = 'RequestContentParseError';
}
