
import { FastifyReply } from 'fastify';
import * as sch from '../utilities/json-schema';

export interface JsonErrorContent {
	code: ErrorCode;
	message: string;
	help?: string;
}

export const error_content_schema = (errors: ErrorCode[]) => sch.obj({
	code: sch.str_enum(errors),
	message: sch.str(),
	help: sch.str('uri'),
}, { required: [ 'code', 'message' ] });

export function send_json_error(res: FastifyReply, error_code: ErrorCode) {
	const error = errors[error_code];
	res.status(error.status);
	res.header('content-type', 'application/json; charset=utf-8');

	if (error.help_link) {
		res.header('link', `<${error.help_link}>; rel="help"`);
	}

	return {
		code: error_code,
		message: error.message,
		help: error.help_link,
	};
}

export function send_html_error(res: FastifyReply, error_code: ErrorCode, render_html: (code: ErrorCode, error: ErrorInfo) => string) {
	const error = errors[error_code];
	res.status(error.status);
	res.header('content-type', 'text/html; charset=utf-8');
	
	if (error.help_link) {
		res.header('link', `<${error.help_link}>; rel="help"`);
	}

	return render_html(error_code, error);
}

export interface ErrorInfo {
	status: number;
	message: string;
	help_link?: string;
}

export type ErrorCode
	= 'oidc_no_pkce_code_verifier'
	| 'oidc_callback_params_invalid'
	| 'oidc_token_fetch_failed'
	| 'oidc_userinfo_fetch_failed'
	;

const errors: Record<ErrorCode, ErrorInfo> = Object.freeze({
	oidc_no_pkce_code_verifier: {
		status: 401,
		message: 'No PKCE code verifier provided (are cookies working correctly?)',
	},
	oidc_callback_params_invalid: {
		status: 401,
		message: 'Login callback parameters invalid',
	},
	oidc_token_fetch_failed: {
		status: 401,
		message: 'Failed to fetch token set from OpenID Connect provider',
	},
	oidc_userinfo_fetch_failed: {
		status: 401,
		message: 'Failed to fetch userinfo from OpenID Connect provider',
	},
});
