
import { FastifyReply } from 'fastify';
import { BaseClient, CallbackParamsType, Issuer, TokenSet, TypeOfGenericClient } from 'openid-client';
import { PKCECookieConfig } from './pkce-cookie';
import { redirect_302_found } from '../http/redirects';
import pino from 'pino';

const scopes = 'openid profile email';

export interface OIDCConfig {
	server_url: string;
	signing_algorithm: 'ES512';
	client_id: string;
	client_secret: string;
	pkce: PKCECookieConfig;
}

export function validate_oidc_conf(conf: unknown) : asserts conf is OIDCConfig {
	// todo: validate config
}

export type OIDCProvider = ReturnType<typeof create_oidc_provider>;

export function create_oidc_provider(conf: OIDCConfig, logger: pino.Logger) {
	const log = logger.child({ logger: 'oidc' });

	let issuer: Issuer;
	let Client: TypeOfGenericClient<BaseClient>;
	let client: BaseClient;

	const ready = (async () => {
		issuer = await Issuer.discover(conf.server_url);
		Client = issuer.Client;
		client = new Client({
			client_id: conf.client_id,
			client_secret: conf.client_secret,
			id_token_signed_response_alg: conf.signing_algorithm,
			authorization_signed_response_alg: conf.signing_algorithm,
		});
	})();

	return {
		get ready() {
			return ready;
		},

		async redirect_to_auth_endpoint(res: FastifyReply, code_challenge: string, redirect_uri: string) {
			await ready;
	
			const uri = client.authorizationUrl({
				response_type: 'code',
				scope: scopes,
				redirect_uri,
				code_challenge,
				code_challenge_method: 'S256'
			});
		
			return redirect_302_found(res, uri, 'Logging in with OpenID Connect');
		},

		parse_callback_params(url: string) {
			return client.callbackParams(url);
		},

		async fetch_token_set(callback_uri: string, params: CallbackParamsType, pkce_code_verifier: string) {
			try {
				return await client.callback(callback_uri, params, {
					response_type: 'code',
					code_verifier: pkce_code_verifier
				}, { });
			}

			catch (error) {
				log.error({ error }, 'failed to exchange callback params and pkce verifier for token set from OIDC provider');
				return null;
			}
		},

		async fetch_user_info(token_set: TokenSet) {
			try {
				return await client.userinfo(token_set);
			}

			catch (error) {
				log.error({ error }, 'failed to fetch userinfo with token set from OIDC provider');
				return null;
			}
		},
	};
}
