
## Login Flow

- User lands on the website, not logged in, and is directed to `/login` page
- On the `/login` page, the user selects the "Login with OpenID Connect" button, which sends a POST to `/login`
- The POST to `/login`:
	- Generates a new **PKCE code** using the `node:crypto` built-in `randomBytes` <sup>[[node:crypto randomBytes]]</sup>, and uses SHA256 to create the **PKCE challenge** from it
  - The **PKCE code** is sent back to the client in the response as a cookie, to be sent back for verification later
    - This cookie is `HttpOnly`, **SHOULD** be `Secure` in any production environment (configurable), and is `SameSite=Lax` (so it gets sent to the `/login-callback` request later)
  - Responds with a 302 redirect to the OIDC authorization URL, including the **PKCE challenge** in the params
- The user logins with their OIDC provider, which redirects the user back to `/login-callback` after successful authentication, including **callback parameters** that can be used to fetch tokens
- The `/login-callback` request:
  - Includes the **PKCE code** in the request cookies (set by `/login` above, sent back here for verification)
  - The service makes a request to the OIDC provider with the **callback parameters** and **PKCE code** (which it will verify using SHA256) to retrieve the tokens
  - The service makes a request to the OIDC provider for userinfo using the retrieved token
  - The subject from the userinfo is used to identify the user in the local data store
  - The userinfo from the OIDC provider is stored in the local user profile
  - A session is generated (see **Session Generation Flow** below) for the identified user
  - The generated **session key** is sent to the user in a cookie
    - This cookie is `HttpOnly`, **SHOULD** be `Secure` in any production environment (configurable), and is `SameSite=Strict`
  - Responds with a 200 response containing an HTTP Refresh header and HTML meta tag, both indicating that the client should redirect to the `/` page
    - For example: `Refresh: 0;URL='https://example.com/'` and `<meta http-equiv="refresh" content="0;URL='https://example.com/'"/>`
  



## Session Generation Flow

- Two random values are generated (using the `node:crypto` built-in `randomBytes` <sup>[[node:crypto randomBytes]]</sup>):
  - A random 128-bit **session prefix**
  - A random 384-bit **session secret**
- Both values are base64url encoded <sup>[[RFC 4648.5]]</sup> and concatenated with a period (`.`) separator to form the **session key** (i.e. `<128-bit base64 prefix>.<384-bit base64 secret>`)
  - Example: `NpKxxgBSyM9U8EOT.3-SxT9fxzd2HXDCe9CEPPhZC4uR6CuVBwbl_6Cy_o9oSWVGy`
- The **session secret** is hashed (using argon2id <sup>[[RFC 9106]]</sup>) for server-side storage to be verified later (see **Session Verification Flow** below)
- Additionally, the **session prefix**, local user ID, generation timestamp, and expiration timestamp (based on configurable TTL) are stored alongside the **hashed session secret**
  - The **session prefix** is the "primary key", used for indexed lookups



## Session Verification Flow

- User requests a resource or attempts to perform an action
- The service will look for a **session key** in two places (in priority order):
  - In the `Authorization` request header, in the form of a `Bearer` token
    - Example: `Authorization: Bearer NpKxxgBSyM9U8EOT.3-SxT9fxzd2HXDCe9CEPPhZC4uR6CuVBwbl_6Cy_o9oSWVGy`
  - In a session key cookie (name depends on config)
    - Example: `Cookie: session_key=NpKxxgBSyM9U8EOT.3-SxT9fxzd2HXDCe9CEPPhZC4uR6CuVBwbl_6Cy_o9oSWVGy`
- The **session prefix** from the **session key** is used to locate the correct session data in the data store
- The **session secret** from the **session key** is verified against the **hashed session secret** stored on the session (using argon2id <sup>[[RFC 9106]]</sup>)
- The stored expiration timestamp is compared against the current time to ensure the session is not expired
- The user ID on the session object is used to determine what user is logged in
- If any of the above steps fail, the service will respond with a 401 error





[RFC 4648.5]: <https://www.rfc-editor.org/rfc/rfc4648#section-5>
[RFC 9106]: <https://www.rfc-editor.org/rfc/rfc9106.html>
[node:crypto randomBytes]: <https://nodejs.org/api/crypto.html#cryptorandombytessize-callback>
