
import { pino } from 'pino';
import { rand } from '../utilities/rand';
import { FastifyReply } from 'fastify';
import { Req } from '../http/request';
import { invalidate_cookie, parse_req_cookies, set_cookie } from '../http/cookies';
import type { Argon2HashProvider } from './argon-hash';
import type { SessionData, StorageProvider, UserData } from '../storage';
import { Snowflake } from '../utilities/snowflake-uid';

export interface SessionKey {
	full_key: string;
	raw_key: string;
	prefix: string;
}

export type SessionProvider = ReturnType<typeof create_session_provider>;

export interface SessionCookieConfig {
	name: string;
	secure: boolean;
	ttl: number;
	pepper: string;
}

export function validate_session_cookie_conf(conf: unknown) : asserts conf is SessionCookieConfig {
	// todo: validate config
}

export function create_session_provider(conf: SessionCookieConfig, logger: pino.Logger, argon2: Argon2HashProvider, storage: StorageProvider) {
	const session_logger = logger.child({ logger: 'session' });
	const self = {
		async generate_key() : Promise<SessionKey> {
			const bytes = await rand(48);
			const base64 = bytes.toString('base64url');
			const prefix = base64.slice(0, 16);
			const raw_key = base64.slice(16);
			const full_key = `${prefix}.${raw_key}`;
			
			return { prefix, raw_key, full_key };
		},
		hash_key(key: SessionKey) : Promise<string> {
			return argon2.hash(conf.pepper + key.raw_key);
		},
		parse_key(full_key: string) : SessionKey {
			const [ prefix, raw_key, ...rest ] = full_key.split('.');
		
			if (rest && rest.length) {
				throw new ParseSessionKeyError();
			}
		
			return { prefix, raw_key, full_key };
		},
		verify_key(key: SessionKey, session: SessionData) : Promise<boolean> {
			return argon2.verify(conf.pepper + key.raw_key, session.key_hash);
		},
		write_to_cookie(res: FastifyReply, key: SessionKey) {
			const session_expire = new Date(Date.now() + (conf.ttl * 1000));
			set_cookie(res, conf.name, key.full_key, session_expire, conf.secure, 'Strict');
		},
		reset(res: FastifyReply) {
			invalidate_cookie(res, conf.name, conf.secure);
		},
		async create_session_data(user_id: Snowflake, session_key: SessionKey) : Promise<SessionData> {
			return {
				user_id: user_id,
				prefix: session_key.prefix,
				key_hash: await argon2.hash(session_key.raw_key),
				started: new Date(),
				expires: new Date(Date.now() + (conf.ttl * 1000)),
			};
		},
		async check_login(req: Req, token?: string) {
			const log = session_logger.child({ reqId: req.id }, {
				redact: [
					'session.key_hash'
				],
			});
		
			let full_key: string;
			let key_source: 'header' | 'cookie' | 'given';

			if (token) {
				full_key = token;
				key_source = 'given';
				log.debug('token provided by caller');
			}

			// If an `Authorization` header is present, it take precedence
			// for providing credentials
			else if (req.headers.authorization) {
				key_source = 'header';
				log.debug('found authorization header, checking for bearer token');

				const header = req.headers.authorization;

				if (! header.startsWith('Bearer ')) {
					log.debug('expected authorization header to start with "Bearer ", but it did not');
					throw new MalformedAuthorizationHeaderError();
				}

				full_key = header.slice(7);
				log.debug('found key in authorzation header');
			}

			// Otherwise, assume the user is using the web UI, and will provide
			// the API key in a cookie
			else {
				key_source = 'cookie';
				log.debug('no authorization header found, checking for session cookie');

				const cookies = parse_req_cookies(req);
				const session_cookie = cookies[conf.name];
			
				if (! session_cookie) {
					log.debug('no session cookie found');
					throw new AuthTokenNotProvidedError();
				}

				full_key = session_cookie;
				log.debug('found key in session cookie');
			}

			const key = self.parse_key(full_key);
			const session = await storage.get_session(key.prefix);

			if (! session) {
				log.debug('failed to find valid, matching session in store');
				throw new AuthTokenInvalidError();
			}

			log.debug({ session }, 'found session in store');

			const key_verified = await self.verify_key(key, session);

			if (! key_verified) {
				log.debug('session found, but the key failed verification against key hash');
				throw new AuthTokenInvalidError();
			}

			let user: UserData;

			if (session.user_id) {
				user = await storage.get_user(session.user_id);

				if (! user) {
					log.error('failed to lookup user associated with session (referential integrity violation)');
					throw new SessionInvalidError();
				}
			}

			log.debug('all checks successful, user is logged in');

			req.session = { key, data: session, user };
		},
	};

	return self;
}

export class ParseSessionKeyError extends Error {
	public readonly name = 'ParseSessionKeyError';
	constructor() {
		super('invalid session key');
	}
}

export class MalformedAuthorizationHeaderError extends Error {
	public readonly name = 'MalformedAuthorizationHeaderError';
	constructor() {
		super('malformed authorization header');
	}
}

export class AuthTokenNotProvidedError extends Error {
	public readonly name = 'AuthTokenNotProvidedError';
	constructor() {
		super('auth token not provided');
	}
}

export class AuthTokenInvalidError extends Error {
	public readonly name = 'AuthTokenInvalidError';
	constructor() {
		super('auth token invalid');
	}
}

export class SessionInvalidError extends Error {
	public readonly name = 'SessionInvalidError';
	constructor() {
		super('session invalid');
	}
}
