
import { load_conf, validate_conf } from './conf';

import { create_logger } from './logger';
import { create_oidc_provider } from './security/openid-connect';
import { create_pkce_cookie_provider } from './security/pkce-cookie';
import { create_snowflake_provider } from './utilities/snowflake-uid';
import { create_storage_provider } from './storage';
import { create_argon_hash_provider } from './security/argon-hash';
import { create_session_provider } from './security/session';

import { create_http_metadata_server } from './http-metadata/server';
import { HttpWebDependencies, create_http_web_server } from './http-web/server';

main();

async function main() {
	// Load and validate configuration
	const conf = await load_conf();
	validate_conf(conf);
	
	// Create the logger and storage
	const logger = create_logger(conf.logging);
	const storage = create_storage_provider(conf.storage, logger.child({ logger: 'storage' }));

	// Create the metadata server
	const http_meta = create_http_metadata_server(conf.http_meta, {
		logger: logger.child({ logger: 'http-meta' }),
		ready_check,
		storage,
	});
	await http_meta.listen();

	// Create all of the core feature providers
	const oidc = create_oidc_provider(conf.oidc, logger);
	const pkce_cookie = create_pkce_cookie_provider(conf.pkce_cookie);
	const snowflake = create_snowflake_provider(conf.snowflake_uid);
	const argon2 = create_argon_hash_provider(conf.argon2);
	const session = create_session_provider(conf.session_cookie, logger, argon2, storage);

	// Wait for any async init steps
	await oidc.ready;
	await storage.ready;

	// Perform any cleanup steps before starting up
	await storage.cleanup_old_sessions();

	const deps: HttpWebDependencies = {
		logger: logger.child({ logger: 'http-web' }),
		oidc,
		pkce_cookie,
		snowflake,
		storage,
		argon2,
		session,
	};

	// Create the main web server
	const http_web = create_http_web_server(conf.http_web, deps);
	await http_web.setup_plugins();
	await http_web.listen();

	function ready_check() {
		try {
			const status = http_web.status;
			return status;
		}

		catch (error) {
			return 'unstarted';
		}
	}
}
