
import type { SessionData } from '../provider';
import type { Snowflake } from '../../utilities/snowflake-uid';
import { read_data_file, write_data_file, delete_data_file, read_data_dir } from '../files';

export async function create_session(data: SessionData) {
	const json = JSON.stringify(to_json(data), null, '  ');
	await write_data_file(`sessions/${data.prefix}`, json, 'wx');
}

export async function get_session(prefix: string) {
	const data = await read_data_file<SessionJson>(`sessions/${prefix}`, 'json');
	return from_json(data);
}

export async function delete_session(prefix: string) {
	await delete_data_file(`sessions/${prefix}`);
}

export async function cleanup_old_sessions() {
	const files = await read_data_dir('sessions');
	const promises = files.map((file) => check_session_for_expiration(file));
	await Promise.all(promises);
}

async function check_session_for_expiration(prefix: string) {
	let session: SessionData;

	try {
		session = await get_session(prefix);
	}

	catch (error) {
		// todo: if the session doesn't exist, skip
		throw error;
	}

	if (session.expires.getTime() <= Date.now()) {
		await delete_session(prefix);
	}
}

interface SessionJson {
	prefix: string;
	key_hash: string;
	user_id: Snowflake;
	started: string;
	expires: string;
}

function to_json(session: SessionData) : SessionJson {
	return {
		prefix: session.prefix,
		key_hash: session.key_hash,
		user_id: session.user_id,
		started: session.started.toISOString(),
		expires: session.expires.toISOString(),
	};
}

function from_json(json: SessionJson) : SessionData {
	return {
		prefix: json.prefix,
		key_hash: json.key_hash,
		user_id: json.user_id,
		started: new Date(Date.parse(json.started)),
		expires: new Date(Date.parse(json.expires)),
	};
}
