
import { DB } from './db';
import { pino } from 'pino';
import { Pool } from 'generic-pool';
import { StorageProvider } from '../provider';
import { create_sqlite_pool } from './pool';
import { SQLite3StorageConfig } from './config';
import { bind_first_param } from '../../utilities/bind';

import { create_session, get_session, delete_session, cleanup_old_sessions } from './v1/sessions';
import { get_user, create_user, update_user, delete_user, get_user_by_oidc_sub } from './v1/users';
import { bring_db_schema_up_to_date } from './migrate';

export function create_sqlite3_storage_provider(conf: SQLite3StorageConfig, logger: pino.Logger) : StorageProvider {
	let status: 'ok' | 'warning' | 'updating' | 'unavailable' = 'unavailable';

	const db: DB = {
		conf,
		logger,
		settings: null,
		users: null,
		sessions: null,
	};

	db.settings = create_sqlite_pool('settings.db', db);
	db.users = create_sqlite_pool('users.db', db);
	db.sessions = create_sqlite_pool('sessions.db', db);

	const ready = bring_db_schema_up_to_date(db).then(() => {
		status = 'ok';
	});

	return {
		ready,

		status() {
			return {
				engine: 'sqlite3',
				status: status,
				pools: {
					settings: pool_stats(db.settings),
					users: pool_stats(db.users),
					sessions: pool_stats(db.sessions),
				},
			};
		},
		
		// Login Sessions
		get_session: bind_first_param(get_session, db),
		create_session: bind_first_param(create_session, db),
		delete_session: bind_first_param(delete_session, db),
		cleanup_old_sessions: bind_first_param(cleanup_old_sessions, db),

		// Users
		get_user: bind_first_param(get_user, db),
		get_user_by_oidc_sub: bind_first_param(get_user_by_oidc_sub, db),
		create_user: bind_first_param(create_user, db),
		update_user: bind_first_param(update_user, db),
		delete_user: bind_first_param(delete_user, db),
	};

	function pool_stats(pool: Pool<any>) {
		return {
			min: pool.min,
			max: pool.max,
			available: pool.available,
			borrowed: pool.borrowed,
			pending: pool.pending,
		};
	}
}
