
import type { DB } from './db';
import type { ISOTimestamp } from '../../utilities/types';
import type * as sqlite3 from 'sqlite3';

import { get_one, run } from './run';
import { type } from './schema-utils';
import { migrate_v0_to_v1 } from './v1/migrate';

const supported_db_version = 1;

export let updating = false;

interface Migration {
	(db: DB): Promise<void>;
}

/**
 * Ensures that the database files exist and are updated to their latest schema versions
 */
export async function bring_db_schema_up_to_date(db: DB, no_update = false) {
	updating = true;
	const log = db.logger.child({ source: 'storage/sqlite3/migrate' });
	const conn = await db.settings.acquire();

	log.info(`checking for any needed db updates`);

	await begin_transaction(db, conn);
	await create_table_migration_history(db, conn);
	
	let { version: db_version } = await get_current_schema_version(db, conn) || { version: 0 };

	if (db_version === supported_db_version) {
		log.info(`db already up to date`);
		await commit(db, conn);
		await db.settings.release(conn);
		updating = false;
		return;
	}

	if (db_version > supported_db_version || db_version < 0) {
		console.error('settings.db seems to contain an unknown or unsupported version');
		await commit(db, conn);
		await db.settings.release(conn);
		process.exit(1);
	}

	if (no_update) {
		console.error('settings.db is out of date, but no_update flag is set');
		await commit(db, conn);
		await db.settings.release(conn);
		process.exit(1);
	}
	
	log.info(`db needs updates`, { db_version, new_version: supported_db_version });

	while (db_version < supported_db_version) {
		const now = (new Date).toISOString() as ISOTimestamp;
		log.info(`performing update from v${db_version} to v${db_version + 1}...`);
		await migrations[db_version](db);
		db_version++;
		await append_migration(db, conn, db_version, now);
	}

	await commit(db, conn);
	await db.settings.release(conn);
	updating = false;
}

const migrations: Migration[] = [
	migrate_v0_to_v1,
	// migrate_v1_to_v2,
];







// ===== Migration History Table =====

const sql_create_table_migration_history = `
create table if not exists migration_history (
	id ${type.rowid} primary key,
	version int not null,
	migrated_at datetime not null
)`;

async function create_table_migration_history(db: DB, conn: sqlite3.Database) {
	await run(db, conn, sql_create_table_migration_history);
}

const sql_begin_transaction = `
begin transaction
`;

async function begin_transaction(db: DB, conn: sqlite3.Database) {
	await run(db, conn, sql_begin_transaction);
}

const sql_commit = `
commit
`;

async function commit(db: DB, conn: sqlite3.Database) {
	await run(db, conn, sql_commit);
}

const sql_get_current_schema_version = `
select
	hist.version as version
from migration_history hist
order by id desc
limit 1
`;

function get_current_schema_version(db: DB, conn: sqlite3.Database) {
	return get_one<{ version: number }>(db, conn, sql_get_current_schema_version);
}

const sql_append_migration = `
insert into migration_history
	(version, migrated_at)
values
	(?, ?)
`;

async function append_migration(db: DB, conn: sqlite3.Database, version: number, migrated_at: ISOTimestamp) {
	await run(db, conn, sql_append_migration, [ version, migrated_at ]);
}
