
import { DB } from './db';
import { run } from './run';
import { data_path } from '../files';
import { promises as fs } from 'fs';
import { createPool, Factory, Pool } from 'generic-pool';
import * as sqlite3 from 'sqlite3';

let next_conn_id = 1;

export const db_info_map = new Map<sqlite3.Database, DBInfo>();

export type DBPool = Pool<sqlite3.Database>;

export interface DBInfo {
	file: string;
	mode: number;
	conn_id: number;
}

export function create_sqlite_pool(file: string, db: DB) : DBPool {
	const path = data_path(file);
	const mode = sqlite3.OPEN_CREATE | sqlite3.OPEN_READWRITE;
	const factory: Factory<sqlite3.Database> = {
		create() {
			return open(db, path, mode);
		},
		destroy(conn: sqlite3.Database) {
			return close(db, conn);
		},
	};

	const pool = createPool(factory, {
		min: db.conf.pool_min,
		max: db.conf.pool_max,
	});

	return pool;
}

function open(db: DB, file: string, mode: number) : Promise<sqlite3.Database> {
	const conn_id = next_conn_id++;
	const db_info: DBInfo = { file, mode, conn_id };
	const log = db.logger.child(db_info);
	
	log.info('opening database connection');
	
	return new Promise<sqlite3.Database>((resolve, reject) => {
		const conn = new sqlite3.Database(file, mode, async (error) => {
			if (error) {
				log.error(error);
				return reject(error);
			}

			db_info_map.set(conn, db_info);

			log.debug('database open');
			log.debug('confirming file permissions = 0600');
			
			// Ensure the file is not accessible to anyone but the server user
			await fs.chmod(file, 0o600);

			log.debug('enabling foreign keys');

			// Enable foreign keys
			await run(db, conn, 'PRAGMA foreign_keys = ON');

			log.info('ready');
			resolve(conn);
		});
	});
}

function close(db: DB, conn: sqlite3.Database) : Promise<void> {
	const db_info = db_info_map.get(conn);
	const log = db.logger.child(db_info);

	log.info('closing database connection');

	return new Promise<void>((resolve, reject) => {
		conn.close((error) => {
			if (error) {
				return reject(error);
			}

			log.info('closed');
			resolve();
		});
	});
}

