
import { DB } from './db';
import { db_info_map } from './pool';
import * as sqlite3 from 'sqlite3';

let next_query_id = 1;

export function run(db: DB, conn: sqlite3.Database, query: string, params: any[] | object = [ ]) {
	const query_id = next_query_id++;
	const info = Object.assign({ query_id }, db_info_map.get(conn));
	const log = db.logger.child(info);
	const start = Date.now();

	query = sql(query);
	
	return new Promise<sqlite3.RunResult>((resolve, reject) => {
		log.trace(`run: ${query}`);

		conn.run(query, params, function(error) {
			if (error) {
				log.error('run failed: ' + error.stack);
				return reject(error);
			}

			log.debug({ duration: Date.now() - start }, `run complete: ${trunc(query)}`);
			resolve(this);
		});
	});
}

export function get_one<T>(db: DB, conn: sqlite3.Database, query: string, params: any[] | object = [ ]) {
	const query_id = next_query_id++;
	const info = Object.assign({ query_id }, db_info_map.get(conn));
	const log = db.logger.child(info);
	const start = Date.now();
	
	query = sql(query);

	return new Promise<T>((resolve, reject) => {
		log.trace(`get_one: ${query}`);

		conn.get(query, params, function(error, row) {
			if (error) {
				log.error('get_one failed: ' + error.stack);
				return reject(error);
			}

			log.debug({ duration: Date.now() - start }, `get_one complete: ${trunc(query)}`);
			resolve(row as any);
		});
	});
}

export function get_all<T>(db: DB, conn: sqlite3.Database, query: string, params: any[] | object = [ ]) {
	const query_id = next_query_id++;
	const info = Object.assign({ query_id }, db_info_map.get(conn));
	const log = db.logger.child(info);
	const start = Date.now();

	query = sql(query);
	
	return new Promise<T[]>((resolve, reject) => {
		log.trace(`get_all: ${query}`);

		conn.all(query, params, function(error, rows) {
			if (error) {
				log.error('get_all failed: ' + error.stack);
				return reject(error);
			}

			log.debug({ duration: Date.now() - start }, `get_all complete: ${trunc(query)}`);
			resolve(rows as any[]);
		});
	});
}

function trunc(sql: string) {
	return `${sql.slice(0, 20)}...`;
}

export function sql(query: string) {
	return query.replace(/\n/g, ' ').replace(/\s+/g, ' ').trim();
}
