
import type { DB } from '../db';
import type { SessionData } from '../../provider';
import type { ISOTimestamp } from '../../../utilities/types';
import type { Snowflake } from '../../../utilities/snowflake-uid';

import { get_one, run } from '../run';

export interface SessionRow {
	prefix: string;
	key_hash: string;
	user_id: Snowflake;
	started: ISOTimestamp;
	expires: ISOTimestamp;
}

const sql_get_session = `
select
	session.prefix as prefix,
	session.key_hash as key_hash,
	session.user_id as user_id,
	session.started as started,
	session.expires as expires
from sessions session
`;

export async function get_session(db: DB, prefix: string) {
	const conn = await db.sessions.acquire();

	try {
		const data = await get_one<SessionRow>(db, conn, sql_get_session + 'where session.prefix = ?', [ prefix ]);

		if (! data) {
			return null;
		}
		
		return from_db_row(data);
	}

	finally {
		db.sessions.release(conn);
	}
}

const sql_create_session = `
insert into sessions
	(prefix, key_hash, user_id, started, expires)
values
	(?, ?, ?, ?, ?)
`;

export async function create_session(db: DB, data: SessionData) {
	const conn = await db.sessions.acquire();

	try {
		await run(db, conn, sql_create_session, [
			data.prefix,
			data.key_hash,
			data.user_id,
			data.started,
			data.expires,
		]);
	}

	finally {
		db.sessions.release(conn);
	}
}

const sql_delete_session = `
delete from sessions
where prefix = ?
`;

export async function delete_session(db: DB, prefix: string) {
	const conn = await db.sessions.acquire();

	try {
		await run(db, conn, sql_delete_session, [ prefix ]);
	}

	finally {
		db.sessions.release(conn);
	}
}

const sql_cleanup_old_sessions = `
delete from sessions
where expires <= ?
`;

export async function cleanup_old_sessions(db: DB) {
	const now = (new Date).toISOString();
	const conn = await db.sessions.acquire();

	try {
		await run(db, conn, sql_cleanup_old_sessions, [ now ]);
	}

	finally {
		db.sessions.release(conn);
	}
}

function from_db_row(row: SessionRow) : SessionData {
	return {
		prefix: row.prefix,
		key_hash: row.key_hash,
		user_id: row.user_id,
		started: new Date(Date.parse(row.started)),
		expires: new Date(Date.parse(row.expires)),
	};
}
