
import type { DB } from '../db';
import type { UserData } from '../../provider';
import type { Snowflake } from '../../../utilities/snowflake-uid';
import type { HttpURL, Locale, Timezone } from '../../../utilities/types';

import { get_one, run } from '../run';

export interface UserRow {
	id: Snowflake;
	username: string;
	oidc_subject?: string;
	name?: string;
	website?: HttpURL;
	profile?: HttpURL;
	picture?: HttpURL;
	locale: Locale;
	timezone: Timezone;
}

const sql_get_user = `
select
	user.id as id,
	user.username as username,
	user.oidc_subject as oidc_subject,
	user.name as name,
	user.website as website,
	user.profile as profile,
	user.picture as picture,
	user.locale as locale,
	user.timezone as timezone
from users user
`;

export async function get_user(db: DB, user_id: Snowflake) : Promise<UserData> {
	const conn = await db.users.acquire();

	try {
		const data = await get_one<UserRow>(db, conn, sql_get_user + 'where user.id = ?', [ user_id ]);

		if (! data) {
			return null;
		}

		return from_db_row(data);
	}

	finally {
		db.users.release(conn);
	}
}

export async function get_user_by_oidc_sub(db: DB, user_sub: string) : Promise<UserData> {
	const conn = await db.users.acquire();

	try {
		const data = await get_one<UserRow>(db, conn, sql_get_user + 'where user.oidc_subject = ?', [ user_sub ]);

		if (! data) {
			return null;
		}
		
		return from_db_row(data);
	}

	finally {
		db.users.release(conn);
	}
}

const sql_create_user = `
insert into users
	(id, username, oidc_subject, name, website, profile, picture, locale, timezone)
values
	(?, ?, ?, ?, ?, ?, ?, ?, ?)
`;

export async function create_user(db: DB, data: UserData) {
	const conn = await db.users.acquire();

	try {
		await run(db, conn, sql_create_user, [
			data.id,
			data.username,
			data.oidc_subject,
			data.name,
			data.website,
			data.profile,
			data.picture,
			data.locale,
			data.timezone,
		]);
	}

	finally {
		db.users.release(conn);
	}
}

export async function update_user(db: DB, user_id: Snowflake, data: UserData) {
	// todo: update_user
}

const sql_delete_user = `
delete from users
where id = ?
`;

export async function delete_user(db: DB, user_id: Snowflake) {
	const conn = await db.users.acquire();

	try {
		await run(db, conn, sql_delete_user, [ user_id ]);
	}

	finally {
		db.users.release(conn);
	}
}

function from_db_row(json: UserRow) : UserData {
	return {
		id: json.id,
		username: json.username,
		oidc_subject: json.oidc_subject,
		name: json.name,
		locale: json.locale,
		timezone: json.timezone,
		picture: json.picture,
		profile: json.profile,
		website: json.website,
	};
}
