
import type { Func, Params } from './types';

export function simple_memo<T>(ttl: number, func: Func<T>) : Func<T> {
	let value: T;
	let expires: number;

	return function memoized() : T {
		const now = Date.now();

		if (value == null || expires < now) {
			value = func();
			expires = now + ttl;
		}

		return value;
	};
}

export type MemoFunc<T extends Func> = T & {
	invalidate(key: string): (() => void);
	revalidate?(key: string): (() => void);
};

export interface Validator<T extends Func> {
	(args: Params<T>, stored_ms: number, value: ReturnType<T>): boolean | Promise<boolean>;
}

export interface MemoParams<T extends Func> {
	validator?: Validator<T>;
}

export function memo<T extends Func>(ttl: number, func: T, opts: MemoParams<T> = { }) : MemoFunc<T> {
	const cache: Record<string, ReturnType<T>> = Object.create(null);

	func_with_memo.invalidate = invalidate;
	func_with_memo.revalidate = opts.validator ? revalidate : null;

	function func_with_memo(...args: Params<T>) : ReturnType<T> {
		const key = memo_key(args);

		if (cache[key]) {
			return cache[key];
		}

		const now = Date.now();
		const result = func(...args);
		cache[key] = result;
		set_expire(key, args, now);
		return result;
	}

	function set_expire(key: string, args: Params<T>, stored: number) {
		setTimeout(opts.validator ? revalidate(key, args, stored) : invalidate(key), ttl);
	}

	function revalidate(key: string, args: Params<T>, stored: number) {
		return async () => {
			if (await opts.validator(args, stored, cache[key])) {
				set_expire(key, args, stored);
			}

			else {
				invalidate(key);
			}
		};
	}

	function invalidate(key: string) {
		return () => {
			delete cache[key];
		};
	}

	return func_with_memo as unknown as MemoFunc<T>;
}

export function memo_key(args: any[]) : string {
	return args.map((arg) => JSON.stringify(arg)).join('\0\0');
}
